/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mount.h>

#include "fsarchiver.h"
#include "dico.h"
#include "common.h"
#include "fs_ntfs.h"
#include "uuid.h"
#include "filesys.h"

int ntfs_mkfs(struct s_dico *d, char *partition, char *filesystem)
{
	char command[2048];
	char buffer[2048];
	char options[2048];
	char temp[1024];
	u32 temp32;
	u16 temp16;
	
	// there is no option that just displays the version and return 0 in mkfs.ntfs
	if (exec_command(command, sizeof(command), NULL, 0, NULL, 0, "which mkfs.ntfs")!=0)
	{	errprintf("mkfs.btrfs not found. please install ntfs-3g 2009.1.1AR.1 or more recent on your system or check the PATH.\n");
		return -1;
	}
	
	// ---- set the advanced filesystem settings from the dico
	memset(options, 0, sizeof(options));
	if (dico_get_string(d, 0, FSYSHEADKEY_FSLABEL, buffer, sizeof(buffer))==0 && strlen(buffer)>0)
	{	snprintf(temp, sizeof(temp), " --label %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(MSG_VERB2, "FSYSHEADKEY_FSLABEL: mkfs_options+=[%s]\n", temp);
	}
	
	if (dico_get_u16(d, 0, FSYSHEADKEY_NTFSSECTORSIZE, &temp16)==0)
	{	snprintf(temp, sizeof(temp), " -s %ld ", (long)temp16);
		bufcat(options, temp, sizeof(options));
		msgprintf(MSG_VERB2, "FSYSHEADKEY_NTFSSECTORSIZE: mkfs_options+=[%s]\n", temp);
	}
	
	if (dico_get_u32(d, 0, FSYSHEADKEY_NTFSCLUSTERSIZE, &temp32)==0)
	{	snprintf(temp, sizeof(temp), " -c %ld ", (long)temp32);
		bufcat(options, temp, sizeof(options));
		msgprintf(MSG_VERB2, "FSYSHEADKEY_NTFSCLUSTERSIZE: mkfs_options+=[%s]\n", temp);
	}
	
	if (exec_command(command, sizeof(command), NULL, 0, NULL, 0, "mkfs.ntfs -f %s %s", partition, options)!=0)
	{	errprintf("command [%s] failed\n", command);
		return -1;
	}
	
	return 0;
}

int ntfs_read_label(int fd, char *labelbuf, int bufsize, struct s_ntfsinfo *info)
{
	u8 cFileRecord[4096]; // 1024 in most cases
	u64 offset;
	u16 nOffsetSequenceAttribute;
	u8 *cData;
	u32 dwAttribType;
	u32 dwAttribLen;
	bool bAttribResident;	
	u8 *cDataResident;
	u32 nAttrSize;
	u64 i;
	
	// init
	memset(labelbuf, 0, bufsize);
	
	// 1. read $Volume record
	offset = ((u64) info->lcn_of_mft_data_attrib * info->bytes_per_cluster) + ((u64) (3 * info->file_record_size));
	
	if (lseek(fd, offset, SEEK_SET)!=offset)
	{	errprintf("lseek(fd, offset=%lld, SEEK_SET) failed\n", (long long)offset);
		return -1;
	}
	
	if ((read(fd, cFileRecord, info->file_record_size))!=info->file_record_size)
	{	errprintf("cannot read file record\n");
		return -2;
	}
	
	// -------- decode the MFT File record
	nOffsetSequenceAttribute = le16_to_cpu(*((u16*)(cFileRecord+0x14)));
	cData = cFileRecord + nOffsetSequenceAttribute;
	
	do
	{	// szData points to the beginning of an attribute
		dwAttribType = le32_to_cpu(*((u32*)(cData)));
		dwAttribLen = le32_to_cpu(*((u32*)(cData+4)));
		bAttribResident = le8_to_cpu(*((u8*)(cData+8)))==0;
		
		if(dwAttribType == 0x60) // "volume_name"
		{
			if (bAttribResident == false)
			{	errprintf("$volume_name attribute is not resident\n");
				return -3;
			}
			
			nAttrSize = le16_to_cpu(*((u16*)(cData+0x10)));
			cDataResident = cData+le16_to_cpu(*((u16*)(cData+0x14)));
			
			for (i=0; (i < nAttrSize/2) && (i<bufsize-1); i++)
				labelbuf[i] = cDataResident[2*i];
		}
		cData += dwAttribLen;
	} while (dwAttribType != (u32)-1); // attribute list ends with type -1
	
	return 0;
}

int ntfs_getinfo(struct s_dico *d, char *devname)
{
	struct s_ntfsinfo info;
	char bootsect[16384];
	char label[256];
	int ret=0;
	int fd=-1;
	
	if ((fd=open64(devname, O_RDONLY|O_LARGEFILE))<0)
	{	ret=-1;
		errprintf("cannot open(%s, O_RDONLY)\n", devname);
		goto ntfs_getinfo_return;
	}
	
	if (read(fd, bootsect, sizeof(bootsect))!=sizeof(bootsect))
	{	ret=-1;
		errprintf("cannot read the boot sector on %s\n", devname);
		goto ntfs_getinfo_return;
	}
	
	if (memcmp(bootsect+3, "NTFS", 4) != 0)
	{	ret=-1;
		errprintf("cannot find the ntfs signature on %s\n", devname);
		goto ntfs_getinfo_return;
	}
	
	info.bytes_per_sector = le16_to_cpu(*((u16*)(bootsect+0xB)));
	info.sectors_per_clusters = le8_to_cpu(*((u8*)(bootsect+0xD)));
	info.total_sector_count = le64_to_cpu(*((u64*)(bootsect+0x28)));
	info.lcn_of_mft_data_attrib = le64_to_cpu(*((u64*)(bootsect+0x30)));
	info.bytes_per_cluster = info.bytes_per_sector * info.sectors_per_clusters;
	info.clusters_per_mft_record = *((s8*)(bootsect+0x40));
	
	if (info.clusters_per_mft_record > 0)
		info.file_record_size = info.clusters_per_mft_record * info.bytes_per_cluster;
	else
		info.file_record_size = 1 << (-info.clusters_per_mft_record);
	
	msgprintf(MSG_VERB2, "bytes_per_sector=[%lld]\n", (long long)info.bytes_per_sector);
	msgprintf(MSG_VERB2, "sectors_per_clusters=[%lld]\n", (long long)info.sectors_per_clusters);
	msgprintf(MSG_VERB2, "total_sector_count=[%lld]\n", (long long)info.total_sector_count);
	msgprintf(MSG_VERB2, "clusters_per_mft_record=[%lld]\n", (long long)info.clusters_per_mft_record);
	msgprintf(MSG_VERB2, "lcn_of_mft_data_attrib=[%lld]\n", (long long)info.lcn_of_mft_data_attrib);
	msgprintf(MSG_VERB2, "bytes_per_cluster=[%lld]\n", (long long)info.bytes_per_cluster);
	msgprintf(MSG_VERB2, "lcn_of_mft_data_attrib=[%lld]\n", (long long)info.lcn_of_mft_data_attrib);
	msgprintf(MSG_VERB2, "file_record_size=[%lld]\n", (long long)info.file_record_size);
	
	dico_add_u16(d, 0, FSYSHEADKEY_NTFSSECTORSIZE, info.bytes_per_sector);
	dico_add_u32(d, 0, FSYSHEADKEY_NTFSCLUSTERSIZE, info.bytes_per_cluster);
	
	if (ntfs_read_label(fd, label, sizeof(label), &info)!=0)
		snprintf(label, sizeof(label), "<unknonw>");
	
	msgprintf(MSG_VERB2, "ntfs_label=[%s]\n", label);
	dico_add_string(d, 0, FSYSHEADKEY_FSLABEL, label);
	
	close(fd);
ntfs_getinfo_return:
	return ret;
}

int ntfs_mount(char *partition, char *mntbuf, char *fsbuf, int flags)
{
	char stderrbuf[2048];
	char command[2048];
	char options[1024];
	char temp[1024];
	char delims[]="\n\r";
	char *saveptr;
	char *result;
	int foundar;
	
	// init
	memset(options, 0, sizeof(options));
	
	// check that mount.ntfs-3g is available and that it supports "streams_interface"
	if (exec_command(command, sizeof(command), NULL, 0, NULL, 0, "which mount.ntfs-3g")!=0)
	{	errprintf("mount.ntfs-3g not found. please install ntfs-3g 2009.1.1AR.1 (advanced release)\n"
			"or a newer version on your system or check the PATH.\n");
		return -1;
	}
	
	memset(stderrbuf, 0, sizeof(stderrbuf));
	exec_command(command, sizeof(command), NULL, 0, stderrbuf, sizeof(stderrbuf), "mount.ntfs-3g -h");
	
	// check if there is "AR" (advanced release) written in the ntfs-3g version line
	foundar=false;
	result=strtok_r(stderrbuf, delims, &saveptr);
	while(result != NULL && foundar==false)
	{	if ((memcmp(result, "ntfs-3g", 7)==0) && (strstr(result, "AR")!=NULL))
			foundar=true;
		result = strtok_r(NULL, delims, &saveptr);
	}
	
	if (foundar==false)
	{	errprintf("it seems that the standard mount.ntfs-3g version is installed.\n"
			"fsarchiver requires the advanced release (2009.1.1AR.1 or more recent)\n");
		return -1;
	}
	else
	{	msgprintf(MSG_VERB2, "The advanced release of ntfs-3g has been found\n");
	}
	
	// set mount options
	snprintf(temp, sizeof(temp), " -o streams_interface=xattr ");
	bufcat(options, temp, sizeof(options));
	
	if (flags & MS_RDONLY)
	{	snprintf(temp, sizeof(temp), " -o ro ");
		bufcat(options, temp, sizeof(options));
	}
	
	// ---- set the advanced filesystem settings from the dico
	if (exec_command(command, sizeof(command), NULL, 0, NULL, 0, "mount.ntfs-3g %s %s %s", options, partition, mntbuf)!=0)
	{	errprintf("command [%s] failed, make sure a recent version \n"
			"of the advanced release of ntfs-3g is installed\n", command);
		return -1;
	}
	
	return 0;
}

int ntfs_umount(char *partition, char *mntbuf)
{
	char command[2048];

	if (exec_command(command, sizeof(command), NULL, 0, NULL, 0, "which fusermount")!=0)
	{	errprintf("fusermount not found. please install fuse on your system or check the PATH.\n");
		return -1;
	}
	
	if (exec_command(command, sizeof(command), NULL, 0, NULL, 0, "fusermount -u %s", mntbuf)!=0)
	{	errprintf("cannot unmount [%s]\n", mntbuf);
		return -1;
	}
	
	return 0;
}
