/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <ext2fs/ext2fs.h>
#include <blkid/blkid.h>

#include "showpart.h"
#include "common.h"

int extract_info(blkid_dev dev, char *show[], struct s_blkdev *devinfo)
{
	blkid_tag_iterate iter;
	const char *type, *value;
	struct stat statbuf;
	char temp[1024];
	int fd;
	
	// init
	memset(devinfo, 0, sizeof(struct s_blkdev));
	memset(temp, 0, sizeof(temp));
	
	// check params
	if (!dev)
		return -1;
	
	// set the short device name ("/dev/sda1" or "/dev/dm-1")
	snprintf(temp, sizeof(temp), "%s", blkid_dev_devname(dev));
	if (memcmp(temp, "/dev/", 5)==0)
		snprintf(devinfo->devname, sizeof(devinfo->devname), "%s", temp);
	else
		snprintf(devinfo->devname, sizeof(devinfo->devname), "/dev/%s", temp);
	
	// set the usual device name ("/dev/sda1" or "/dev/mapper/volgroup00-logvol00")
	if (readlink(devinfo->devname, temp, sizeof(temp))>0)
		snprintf(devinfo->longname, sizeof(devinfo->longname), "/dev/%s", temp);
	else
		snprintf(devinfo->longname, sizeof(devinfo->longname), "%s", devinfo->devname);
	
	iter = blkid_tag_iterate_begin(dev);
	while (blkid_tag_next(iter, &type, &value) == 0)
	{
		if (strcmp(type, "LABEL")==0)
			snprintf(devinfo->label, sizeof(devinfo->label), "%s", value);
		else if (strcmp(type, "UUID")==0)
			snprintf(devinfo->uuid, sizeof(devinfo->uuid), "%s", value);
		else if (strcmp(type, "TYPE")==0)
			snprintf(devinfo->fsname, sizeof(devinfo->fsname), "%s", value);
	}
	blkid_tag_iterate_end(iter);
	
	if (stat(devinfo->devname, &statbuf)==0)
	{
		devinfo->rdev=statbuf.st_rdev;
		snprintf(devinfo->major, sizeof(devinfo->major), "%d", major(statbuf.st_rdev));
		snprintf(devinfo->minor, sizeof(devinfo->minor), "%d", minor(statbuf.st_rdev));
	}
	if ((fd=open64(devinfo->devname, O_RDONLY|O_LARGEFILE))>=0)
	{
		devinfo->devsize=lseek64(fd, 0, SEEK_END);
		close(fd);
	}
	return 0; // ok
}

int get_blkid(struct s_blkdev *blkdev, int maxblkdev)
{
	int best; // index of the best item found in old array
	int pos=0; // pos of latest item in new array
	struct s_blkdev blkdev1[FSA_MAX_BLKDEVICES];
	blkid_cache cache = NULL;
	blkid_dev_iterate iter;
	char *show[128] = { NULL, };
	struct s_blkdev tmpdev;
	blkid_dev dev;
	int count=0;
	bool found;
	int i, j;
	
	if (blkid_get_cache(&cache, NULL) < 0)
		return -1;
	blkid_probe_all(cache);
	
	// ---- 1. get all the devices as they are returned by blkid
	iter = blkid_dev_iterate_begin(cache);
	while (blkid_dev_next(iter, &dev)==0 && (count < FSA_MAX_BLKDEVICES) && (count < maxblkdev))
	{
		if ((dev=blkid_verify(cache, dev))!=NULL && extract_info(dev, show, &tmpdev)==0 && strcmp(tmpdev.devname, "/dev/root")!=0)
		{
			// check that this device is not already in the list
			for (found=false, i=0; (i < count) && (found==false); i++)
			{
				if (blkdev1[i].rdev==tmpdev.rdev)
					found=true;
			}
			if (found==false)
				blkdev1[count++]=tmpdev;
		}
	}
	blkid_dev_iterate_end(iter);
	
	// ---- 2. sort the devices
	for (pos=0, i=0; i<count; i++)
	{
		// set best to the first available item in the old array
		for (j=0, best=-1; (j<count) && (best==-1); j++)
			if (blkdev1[j].rdev!=0)
				best=j;
		// find the index of the best item in the old array
		for (j=0; j<count; j++)
			if ((blkdev1[j].rdev > 0) && (blkdev1[j].rdev < blkdev1[best].rdev))
				best=j;
		// move item to the new array
		blkdev[pos++]=blkdev1[best];
		blkdev1[best].rdev=0;
	}
	
	return count;
}

struct s_diskinfo diskinfo[]=
{
	{false,	"[%-16s] ",	"[=====DEVICE=====] "},
	{false,	"[%-11s] ",	"[==FILESYS==] "},
	{false,	"[%-15s] ",	"[=====LABEL=====] "},
	{false,	"[%12s] ",	"[====SIZE====] "},
	{false,	"[%3s] ",	"[MAJ] "},
	{false,	"[%3s] ",	"[MIN] "},
	{true,	"[%-36s] ",	"[==============LONGNAME==============] "},
	{true,	"[%-40s] ",	"[==================UUID==================] "},
	{false,	"",		""}
};

char *get_info(char *bufdat, int bufsize, struct s_blkdev *blkdev, int item)
{
	memset(bufdat, 0, bufsize);
	switch (item)
	{
		case 0:	snprintf(bufdat, bufsize, "%s", blkdev->devname); break;
		case 1:	snprintf(bufdat, bufsize, "%s", blkdev->fsname); break;
		case 2:	snprintf(bufdat, bufsize, "%s", blkdev->label); break;
		case 3:	format_size(blkdev->devsize, bufdat, bufsize, 'h'); break;
		case 4:	snprintf(bufdat, bufsize, "%s", blkdev->major); break;
		case 5:	snprintf(bufdat, bufsize, "%s", blkdev->minor); break;
		case 6:	snprintf(bufdat, bufsize, "%s", blkdev->longname); break;
		case 7:	snprintf(bufdat, bufsize, "%s", blkdev->uuid); break;
	};
	return bufdat;
}

int do_probe(bool details)
{
	struct s_blkdev blkdev[FSA_MAX_BLKDEVICES];
	char temp[1024];
	int res;
	int i, j;
	
	if ((res=get_blkid(blkdev, FSA_MAX_BLKDEVICES))<1)
	{	msgprintf(MSG_FORCE, "No filesystem detected\n");
		return -1;
	}
	
	for (j=0; diskinfo[j].title[0]; j++)
	{
		if (details==true || diskinfo[j].detailed==false)
			msgprintf(MSG_FORCE, "%s", diskinfo[j].title);
	}
	msgprintf(MSG_FORCE, "\n");
	
	for (i=0; i < res; i++)
	{
		for (j=0; diskinfo[j].title[0]; j++)
		{
			if (details==true || diskinfo[j].detailed==false)
				msgprintf(MSG_FORCE, diskinfo[j].format, get_info(temp, sizeof(temp), &blkdev[i], j));
		}
		msgprintf(MSG_FORCE, "\n");
	}
	
	return 0;
}
